// Likelihood evaluator and other functions for -mybipoisson-.

clear all

mata:
mata set matastrict on

// This is the method-lf0 evaluator() -- no derivatives.
void mybipoisson_lf0(transmorphic scalar ML, real scalar    todo,
                     real rowvector      b , real colvector lnfj,
                     real matrix         S , real matrix    H)
{
        real colvector  y1, y2, xi1, xi2, xi0

        y1  = moptimize_util_depvar(ML, 1)
        y2  = moptimize_util_depvar(ML, 2)
        xi1 = exp(moptimize_util_xb(ML, b, 1)) 
        xi2 = exp(moptimize_util_xb(ML, b, 2)) 
        xi0 = J(rows(y1), 1, exp(moptimize_util_xb(ML, b, 3)))

        lnfj = ln( bipoissonp_xi(y1, y2, xi1, xi2, xi0) )
}

// This is the method-lf2() evaluator. You can also use it with
// methods lf0 and lf1.
void mybipoisson_lf2(transmorphic scalar ML, real scalar    todo,
                     real rowvector      b , real colvector lnfj,
                     real matrix         S , real matrix    H)
{
        real colvector  y1, y2, xi1, xi2, xi0
        real matrix     G, Hraw, H11, H12, H13, H22, H23, H33

        y1  = moptimize_util_depvar(ML, 1)
        y2  = moptimize_util_depvar(ML, 2)
        xi1 = exp(moptimize_util_xb(ML, b, 1)) 
        xi2 = exp(moptimize_util_xb(ML, b, 2)) 
        xi0 = J(rows(y1), 1, exp(moptimize_util_xb(ML, b, 3)))

        // The following lines are necessary to handle constant-
        // only models. All of our functions assume xi1 and xi2
        // are colvectors; -ml- turns constant-only equations into
        // scalars.
        if (rows(xi1) == 1) {
                xi1 = J(rows(y1), 1, xi1)
                xi2 = J(rows(y1), 1, xi2)
        }

        lnfj = ln( bipoissonp_xi(y1, y2, xi1, xi2, xi0) )
        if (todo==0) return             // No derivatives 

        G = ln_bip_xi_grad(y1, y2, xi1, xi2, xi0)
        S = J(rows(G), 3, 0)
        S[., 1] = G[.,1] :/ G[.,4]
        S[., 2] = G[.,2] :/ G[.,4]
        S[., 3] = G[.,3] :/ G[.,4]
        if (todo==1) return             // Just gradient

        Hraw = ln_bip_xi_hessian(y1, y2, xi1, xi2, xi0, G)
        H11  = moptimize_util_matsum(ML, 1, 1, Hraw[.,1], lnfj[1])
        H12  = moptimize_util_matsum(ML, 1, 2, Hraw[.,2], lnfj[1])
        H13  = moptimize_util_matsum(ML, 1, 3, Hraw[.,3], lnfj[1])
        H22  = moptimize_util_matsum(ML, 2, 2, Hraw[.,4], lnfj[1])
        H23  = moptimize_util_matsum(ML, 2, 3, Hraw[.,5], lnfj[1])
        H33  = moptimize_util_matsum(ML, 3, 3, Hraw[.,6], lnfj[1])
        H    = ( H11 , H12 , H13 \
                 H12', H22 , H23 \
                 H13', H23', H33 ) 
}

// This version of the bivariate Poisson PMF takes inputs in the
// (mu1, mu2, xi0)-space
real colvector bipoissonp(real colvector y1,
                          real colvector y2,
                          real colvector mu1,
                          real colvector mu2,
                          real colvector xi0) {
        real colvector  xi1, xi2

        // Go from (mu1, mu2, xi0)-space to (xi1, xi2, xi0)-space
        xi1 = mu1 - xi0
        xi2 = mu2 - xi0
        
        return( bipoissonp_xi(y1, y2, xi1, xi2, xi0) )
}

// This version of the bivariate Poisson PMF takes inputs in the
// (xi1, xi2, xi0)-space
real colvector bipoissonp_xi(real colvector y1,
                             real colvector y2,
                             real colvector xi1,
                             real colvector xi2,
                             real colvector xi0)
{
        real scalar     i, lim, p, r
        real colvector  pr

        pr = J(rows(y1), 1, .)
        for(i=1; i<=rows(y1); ++i) {
                lim = min( (y1[i], y2[i]) )
                p = 0
                for(r=0; r<=lim; ++r) {
                        p = p + poissonp(xi0[i], r) *
                                poissonp(xi1[i], y1[i]-r) *
                                poissonp(xi2[i], y2[i]-r) 
                }
                pr[i] = p
        }
        return(pr)
}

// In the book we define Srj = Pr(y1-r)*Pr(y2-r)*Pr(r)
//                  and   Sj = Sum_r Srj
// We don't do the final division by S so we can use these
// ingredients in our Hessian calculations.
real matrix ln_bip_xi_grad(real colvector y1,
                           real colvector y2,
                           real colvector xi1,
                           real colvector xi2,
                           real colvector xi0)
{
        real scalar     j, r, Rj, S, Sr, dSdt1, dSdt2, dSdt3
        real matrix     G

        G = J(rows(y1), 4, 0)   // sic
        for(j=1; j<=rows(y1); ++j) {
                Rj     = min( (y1[j], y2[j]) )
                S      = 0
                dSdt1  = 0
                dSdt2  = 0
                dSdt3  = 0
                for(r=0; r<=Rj; ++r) {
                        Sr = poissonp(xi0[j], r) *
                             poissonp(xi1[j], y1[j]-r) *
                             poissonp(xi2[j], y2[j]-r)
                        S  = S + Sr
                        dSdt1 = dSdt1 + Sr*((y1[j]-r) - xi1[j]) 
                        dSdt2 = dSdt2 + Sr*((y2[j]-r) - xi2[j]) 
                        dSdt3 = dSdt3 + Sr*(r - xi0[j])    // sic
                }
                G[j, 1] = dSdt1
                G[j, 2] = dSdt2
                G[j, 3] = dSdt3
                G[j, 4] = S
        }
        return(G)
}

// Returns observation-level hessian of log-likelihood
// [ d2l/dt1dt1 d2l/dt1dt2 d2l/dt1dt3 d2l/dt2dt2 d2l/dt2dt2 d2l/dt3dt3 ]
// In the book we define Srj = Pr(y1-r)*Pr(y2-r)*Pr(r)
//                  and   Sj = Sum_r Srj
// The matrix G passed in has in the first three columns
// sum_r dSrj / dtk (k = 1,2,3). 
// The fourth column has sum_r Srj = Sj.
real matrix ln_bip_xi_hessian(real colvector y1,
                              real colvector y2,
                              real colvector xi1,
                              real colvector xi2,
                              real colvector xi0,
                              real matrix    G)
{
        real scalar     j, k, m, r, Rj, Srj, u1, u2, u3,
                        S11, S12, S13, S22, S23, S33
        real matrix     Hraw

        Hraw = J(rows(y1), 6, 0)
        // j indexes observations
        for(j=1; j<=rows(y1); ++j) {
                // Get the cross-partial term 
                Rj = min( (y1[j], y2[j]) )
                Srj = 0
                S11 = 0
                S12 = 0
                S13 = 0
                S22 = 0
                S23 = 0
                S33 = 0
                for(r=0; r<=Rj; ++r) {
                        Srj = poissonp(xi0[j], r) *
                              poissonp(xi1[j], y1[j]-r) *
                              poissonp(xi2[j], y2[j]-r)
                        u1  = (y1[j] - r - xi1[j])
                        u2  = (y2[j] - r - xi2[j])
                        u3  = (r - xi0[j])
                        S11 = S11 + Srj*(u1^2 - xi1[j])
                        S12 = S12 + Srj*u1*u2
                        S13 = S13 + Srj*u1*u3
                        S22 = S22 + Srj*(u2^2 - xi2[j])
                        S23 = S23 + Srj*u2*u3
                        S33 = S33 + Srj*(u3^2 - xi0[j])
                }
                Hraw[j, 1] = G[j,4]*S11 - G[j,1]*G[j,1]
                Hraw[j, 2] = G[j,4]*S12 - G[j,1]*G[j,2]
                Hraw[j, 3] = G[j,4]*S13 - G[j,1]*G[j,3]
                Hraw[j, 4] = G[j,4]*S22 - G[j,2]*G[j,2]
                Hraw[j, 5] = G[j,4]*S23 - G[j,2]*G[j,3]
                Hraw[j, 6] = G[j,4]*S33 - G[j,3]*G[j,3]
                Hraw[j, .] = Hraw[j, .] :/ ( G[j,4]^2 )
        }
        return(Hraw)
}



mata mlib create lmybipoisson, replace
mata mlib add lmybipoisson *()
mata mlib index

end
exit
